package e2e_tests

import (
	"fmt"
	"testing"

	"github.com/stretchr/testify/require"

	"gitlab.com/gitlab-org/step-runner/pkg/testutil"
	"gitlab.com/gitlab-org/step-runner/pkg/testutil/bldr"
)

func TestCanFetchAPublishedImage(t *testing.T) {
	registry := bldr.StartOCIRegistryServer(t)

	baseDir := bldr.Files(t).
		WriteFile("step.yml", "spec:\n---\nexec:\n  command: [cat, '${{step_dir}}/app/files/templates/message']").
		WriteFile("files/templates_dir/message", "Hello, World!").
		Build()

	template := `
spec:
---
run:
  - name: publish_image
    step: dist://oci/publish
    inputs:
      registry: %s
      repository: my-image
      tag: 1.0.2
      common:
        files:
          %s/step.yml: step.yml
      platforms:
        %s/%s:  
          files:
            %s/files/templates_dir: /app/files/templates
  - name: run_published_step
    step:
      oci:
        registry: %s
        repository: my-image
        tag: "1"
  - name: echo_registry
    script: "echo reg: ${{steps.publish_image.outputs.registry}}"
  - name: echo_repository
    script: "echo repo: ${{steps.publish_image.outputs.repository}}"
  - name: echo_tag
    script: "echo tag: ${{steps.publish_image.outputs.tag}}"
  - name: echo_ref
    script: "echo ref: ${{steps.publish_image.outputs.ref}}"
  - name: echo_algorithm
    script: "echo algorithm: ${{steps.publish_image.outputs.digest.algorithm}}"
  - name: echo_hash
    script: "echo hash: ${{steps.publish_image.outputs.digest.hash}}"
  - name: echo_digest
    script: "echo digest: ${{steps.publish_image.outputs.digest.value}}"`

	platform := bldr.OCIPlatform.ThisPlatform
	registryAddr := registry.Address()
	testStep := fmt.Sprintf(template, registryAddr, baseDir, platform.OS, platform.Architecture, baseDir, registryAddr)

	_, logs, err := testutil.StepRunner(t).WithDebugLogs().Run(testStep)
	require.NoError(t, err)
	require.Contains(t, logs, "Hello, World!")
	require.Contains(t, logs, `Running step "publish_image"`)
	require.Regexp(t, `INFO published step image=.*/my-image:1.0.2`, logs)
	require.Contains(t, logs, `Running step "run_published_step"`)
	require.Contains(t, logs, "Hello, World!")
	require.Contains(t, logs, "reg: "+registryAddr)
	require.Contains(t, logs, "repo: my-image")
	require.Contains(t, logs, "tag: 1")
	require.Contains(t, logs, "ref: 1")
	require.Regexp(t, `ref: .*/my-image:1.0.2`, logs)
	require.Contains(t, logs, "algorithm: sha256")
	require.Regexp(t, `hash: [0-9a-f]{64}`, logs)
	require.Regexp(t, `digest: sha256:[0-9a-f]{64}`, logs)
}
